@@ -14,6 +14,8 @@ class Node:
     # The recursive differentiation function 
     # Source (Adapted From): https://rufflewind.com/2016-12-30/reverse-mode-automatic-differentiation
     def grad(self):
+        # Note a non-Node type object tries to access .grad(), it will automatically throw an Attribute Error
+        
         # recurse only if the value is not yet cached
         if self.grad_value is None:
             # recursively calculate derivative (adjoint) using chain rule
@@ -23,13 +25,13 @@ class Node:
 
 
 
-
     ### MATHEMATICAL OPERATORS AND BASIC OPERATIONS ###
     # 1) Multiplication, 2) Division, 3) Addition, 4) Subtraction, 5) Power, 6) Arithmetic Negation
     
     # 1) Multiplication
     def __mul__(self, other):
         out = Node()
+        # Assume `other` is a Node object
         try:
             out.val = self.val*other.val
             self.children.append((other.val, out)) # weight = ∂z/∂self = other.value
@@ -157,7 +159,7 @@ class Node:
     def __neg__(self):    
         out = Node()
         try:
-            out.val = 0 - self.val
+            out.val = - self.val
             self.children.append((-1, out)) # weight = ∂z/∂self = -1
         except AttributeError:
             raise TypeError('Negation failed: `input` not an `Node` object or a real number.')
@@ -212,7 +214,11 @@ def log(a, base = None):
     out = Node()
     if base == None:
         try:
-            out.val = np.log(a.val)
+            if a.val == 0:
+                out.val = -np.inf
+                Print('Logarithm computed on 0; negative infinity returned.')
+            else:
+                out.val = np.log(a.val)
             a.children.append((1/a.val, out)) # weight = ∂z/∂self = 1 / self
         except AttributeError:
             try:
@@ -362,19 +368,26 @@ def tanh(a):
 
 # 12) Square Root
 def sqrt(a):
+    out = Node()
     try:
-        return a**0.5
+        out.val = a.val**0.5
+        a.children.append((1./(2*a.val**0.5), out))
     except AttributeError:
         try:
-            return a.val**0.5
+            out.val = a**0.5
         except Exception as e:
             raise TypeError('Square root failed: `input` not an `Node` object or a real number.')   
+    return out
 
+# 13) Logistic Sigmoid 
 def expit(a):
+    out = Node()
     try:
-        return 1/(1+exp(-a))
+        out.val = 1/(1+np.exp(-a.val))
+        a.children.append((1/(1+np.exp(-a.val)), out))
     except AttributeError:
         try:
-            return 1/(1+exp(-a.val))
+            return 1/(1+np.exp(-a))
         except Exception as e:
-            raise TypeError('Logistic Sigmoid failed: `input` not an `Node` object or a real number.')
\ No newline at end of file
+            raise TypeError('Logistic Sigmoid failed: `input` not an `Node` object or a real number.')
+    return out
